
# Helper function making reading external override easier
def parse_value_override(arg):
    i = arg.index("=")
    key = arg[:i].strip()
    val = arg[i+1:].strip()
    return {key:val}


def format_flyway(input_data):
    
    def r_formatter(d, leading=""):
        formatted = []
        for key,value in d.items():

            full_key = ".".join([x for x in [leading, key] if x.strip() != ""])
            if value.__class__ == dict:
                formatted += r_formatter(value, full_key)
            else:
                p = "{} = {}".format(full_key, value )
                formatted.append(p)
        return formatted

    return "\n".join(r_formatter(input_data))






from .exceptions import InvalidProjectStructure, InvalidInputFiles
import os
import itertools
from ecs import render

def verify_input_file(p):
    # path does not have even a value
    if not p:
        raise FileNotFoundError

    # p has a value but it is not a path
    if not os.path.exists(p):
        raise FileNotFoundError

    # If p is a directory and contains no files
    if os.path.isdir(p):
        os.listdir(p)

    # Return value if complies
    return p


def _find_td_ser(project_path, tp="task-definition"):
    """
    Function identifying task-definition within a project directory. The
    exps variable define order/priority for selection of task-definition
    """
    e = [".yml", ".yaml", ".json", "/"]
    exps = ["{}{}".format(tp, x) for x in e]

    # Generate paths
    paths = [os.path.join(project_path, x) for x in exps]

    # Find the first matching file/dir
    for p in paths:
        try:
           verify_input_file(p)
           return p
        except FileNotFoundError as e:
            pass
    
    # Raise exception if no such file is in the project directory
    raise FileNotFoundError


def _find_values(project_path):
    """
    This function identifies values used for tempalte interpolation within
    the values.yml of values/default.yml/yaml/json file.
    """

    # Define potential sources
    exts = ["yml", "yaml", "json"]
    file_names  = ["values", "values/default"]
    
    # Iterate
    for fn in file_names:
        p = itertools.product([fn], exts)
        paths = [os.path.join(project_path, "{}.{}".format(*x)) for x in p]

        for path in paths:
            try:
                verify_input_file(path)
                return [path]
            except FileNotFoundError as e:
                pass

    # Raise exception if no such file is in the project directory
    raise FileNotFoundError
    

def load_project_files(project_path="ecs/", values=None):
    # Find task-definition.yml/json/yaml or task-definition/
    try:
        td = _find_td_ser(project_path)
    except FileNotFoundError as e:
        raise InvalidProjectStructure("Unable to find task-definition")

    # Find service.yml/json/yaml or service/
    try:
        service = _find_td_ser(project_path, tp="service")
    except FileNotFoundError as e:
        raise InvalidProjectStructure("Unable to find service definition")

    # if values == None then look for values.yml/yaml/json
    if values is None or (values.__class__ == list and len(values) == 0):
        try:
            values = _find_values(project_path)
        except FileNotFoundError as e:
            raise InvalidProjectStructure("Unable to find values for interpolation")
    else:
        # Deal with project files loading
        if values.__class__ != list:
            raise Exception('Invalid input variable values. Expecting list type!')
        
        # Lets validate input list
        ep = [os.path.exists(x) for x in values]
        if not all(ep):
            raise InvalidInputFiles

    # Return clean paths to task-definition and value files
    return (td, service, values)


def _load_and_interpolate_values(values, envs):
    # Load values from files
    rvalues = [render.load_path(x) for x in values]
    rvalues = [item for sublist in rvalues for item in sublist]
    values = render.merge_dicts(rvalues)

    # Override values from envs dict
    if not envs.__class__ is dict:
        raise Exception('Envs variable must be dict!')
    values = render.merge_dicts([values, envs])

    # Append extra variables
    values['region'] = os.environ.get('AWS_DEFAULT_REGION', 'us-east-1')

    # Interpolate variables
    if len(values):
        # Return interpolated values
        return render.interpolate_values(values)
    
    # Return values
    return values


def load_project(project_path, values, envs):
    # Load files
    ftd, fservice, fvalues = load_project_files(project_path, values)

    # Load values
    ivalues = _load_and_interpolate_values(fvalues, envs)
    
    # Load task definition
    td = render.load_path(ftd, ivalues)
    mtd = render.merge_dicts(td)

    # Load service definition
    service = render.load_path(fservice, ivalues)
    mservice = render.merge_dicts(service)

    # Load output definition
    try: 
        output = load_output(project_path, values, envs)
    except FileNotFoundError as e:
        output = ""

    # Return generates descriptors
    return mtd, mservice, output


def load_output(project_path, values, envs):
    # The output file name is given
    fn = os.path.join(project_path, 'output.j2')

    # If exists then load otherwise return blank output
    if verify_input_file(fn):
        ivalues = _load_and_interpolate_values(values, envs)
        p = render.load_path(fn, ivalues, raw=True)
        return '\n'.join(p).strip()
    else:
        return ""


import os
import errno
import magic
import yaml
from jinja2 import Template, Environment, FileSystemLoader
from .extra_filters import get_jira_id, get_slug, filter_dateparser, filter_format, decode_aws_secret


# support function allowing to read content of a file or directory specified by path variable
def load_path(path, ivalues=None, raw=False):

    # test if file even exists
    if not os.path.exists(path):
        raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), path)

    # Load as a directory
    if os.path.isdir(path):
        vars = []
        for f in os.listdir(path):
            try:
                vars += load_path(os.path.join(path, f), ivalues)
            except Exception as e:
                # I particularly dont like this way, but it works..
                # TODO: should be modified, so an dedicated exception is raised
                if "Not a loadable yaml format" in str(e):
                    continue
                raise e
        return vars
    else:
        # print ("reading file")
        mime = magic.from_file(path, mime=True)
        if not mime in ["text/plain", "text/html", "application/json", "application/yaml"]:
            raise Exception("File {} is not in the text/plain format, but {}".format(path, mime))

        # Try to read yaml file
        with open(path, "r") as f:
            if not ivalues is None:
                # Interpolate the loaded file - this is most likely task_definition
                # That should be in Jinja2 format
                env = Environment()
                env.filters['jira_id'] = get_jira_id
                env.filters['slug'] = get_slug
                env.filters['dateparser'] = filter_dateparser
                env.filters['format_timestamp'] = filter_format
                env.filters['decode_aws_secret'] = decode_aws_secret
                env.loader = FileSystemLoader(os.path.dirname(path))
                t = env.from_string(f.read())
                r = t.render(**ivalues)
                if raw:
                    data = r
                else:
                    data = yaml.load(r, Loader=yaml.FullLoader)
            else:
                if raw:
                    data = f.read()
                else:
                    # Simply load the yaml/json - this should be file with values
                    data = yaml.load(f, Loader=yaml.FullLoader)
            if data.__class__.__name__ != 'dict' and not raw:
                raise Exception("Not a loadable yaml format - {}".format(path))
            return [data]


# This function will need to merge all loaded dicts
# At the moment, I will implemented very simple merge without deep nesting
# merging support.
def merge_dicts(list_of_dicts):
    _tmp = {}
    for d in list_of_dicts:
        _tmp = merge(_tmp, d)
    return _tmp


def merge(a, b, path=None):
    "merges b into a"
    if path is None:
        path = []
    for key in b:
        if key in a:
            if isinstance(a[key], dict) and isinstance(b[key], dict):
                merge(a[key], b[key], path + [str(key)])
            elif isinstance(a[key], list) and isinstance(b[key], list):
                a[key] += b[key]
            elif isinstance(a[key], str) and isinstance(b[key], str):
                a[key] = b[key]
            elif isinstance(a[key], int) and isinstance(b[key], int):
                a[key] = b[key]
            elif isinstance(a[key], float) and isinstance(b[key], float):
                a[key] = b[key]
            elif a[key] == b[key]:
                pass  # same leaf value
            else:
                raise Exception('Conflict at %s' % '.'.join(path + [str(key)]))
        else:
            a[key] = b[key]
    return a


# This is just a helping function mimicing Ansible's interpolation within variables
# I do not exepct anybody will ever need more than 5 consecutive interpolations
# if so it means there is a very bad pattern how deployment is configured
def interpolate_values(values):
    td = yaml.dump(values, Dumper=yaml.Dumper)
    for i in range(0, 5):
        template = Template(td)
        td = template.render(**values)
        if not "{{" in td:
            return yaml.load(td, Loader=yaml.FullLoader)
    raise Exception("Too many recusions")


# This function takes task_definition in the form of a dict and uses jinja2 for
# variable interplation using values which is another dict composed of all provided
# variables
def render(task_definition, values):
    td = yaml.dump(task_definition, Dumper=yaml.Dumper)
    template = Template(td)
    td = template.render(interpolate_values(values))
    return yaml.load(td, Loader=yaml.FullLoader)



#!/usr/bin/env python


import click
from ecs import project_loader, ecs
import yaml
import json
from pprint import pprint
import os 
import boto3
import botocore


# Shared options
_shared_options = [
    click.option('-p', '--project-path', 'project_path', show_default=True, default="ecs/"),
    click.option('-v', '--val', 'values', type=click.Path(exists=True), multiple=True),
    click.option('-e', 'envs', type=str, multiple=True)
]


def add_options(options):
    def _add_options(func):
        for option in reversed(options):
            func = option(func)
        return func
    return _add_options


# Definition of group necessary for building arguments
@click.group()
def group(**kwargs):
    pass


def _render_project(**kwargs):
    """ Function used for handling project_path, values, and envs options"""
    # Convert env input into dict
    envs = list(kwargs.get('envs', []))
    envs = dict([x.split('=') for x in envs])

    # Load and interpolate project
    return project_loader.load_project(
        kwargs.get('project_path'),
        list(kwargs.get('values', [])),
        envs)


@group.command()
@add_options(_shared_options)
@click.option('-f', '--format', 'oformat', type=click.Choice(['json', 'yaml']), default='json')
def render(**kwargs):
    """ Interpolate and print project configuration """

    # Interpolate the project
    ld = _render_project(**kwargs)
   
    # Print
    for i in zip(['Task definition', 'Service'], ld):
        # Generate in the output format
        if kwargs['oformat'].lower() == 'yaml':
            g = yaml.dump(i[1])
        if kwargs['oformat'].lower() == 'json':
            g = json.dumps(i[1],indent=2)
        
        # Print
        click.secho("* {}:".format(i[0]), fg='green')
        click.echo(g)
    click.secho("* Output:", fg='green')
    click.echo(ld[2])


@group.command()
@add_options(_shared_options)
@click.option('-f', '--format', 'oformat', type=click.Choice(['json', 'yaml']), default='json')
def output(**kwargs):
    """ Interpolate output.j2 file """
    # Interpolate the project
    ld = _render_project(**kwargs)
   
    # Print the output
    click.echo(ld[2])


@group.command()
@add_options(_shared_options)
@click.option('-td', '--task-definition', 'task_definition', default='task-definition.json')
@click.option('-s', '--service', 'service', default='service.json')
@click.option('-o', '--output', 'output', default='output.txt')
def generate(**kwargs):
    """ Generate configuration files which can be used by aws-cli """

    # Interpolate the project
    ld = _render_project(**kwargs)
   
    # Portion generating output files
    for i in zip([kwargs['task_definition'], kwargs['service'], kwargs['output']], ld):
        # Helper print
        click.secho("Generating {}".format(i[0]))

        # Generate output file
        with open(i[0], "w") as f:
            g = json.dumps(i[1],indent=2)
            f.write(g)


@group.command()
@add_options(_shared_options)
@click.option('--force-task-definition', "force_td", is_flag=True)
@click.option('--fast-redeploy', "fast_redeploy", is_flag=True)
@click.option('--wait', "wait_for_service", is_flag=True)
@click.option('-r', '--region', 'aws_region', show_default=True, default="us-east-1")
def install(**kwargs):
    """ Installs task definition and service to given cluster """

    # Interpolate the project
    td, ts, output = _render_project(**kwargs)
    service_name = ts['serviceName']

    # Install task-definition
    os.environ['AWS_DEFAULT_REGION'] = kwargs['aws_region']
    arn_td = ecs.install_or_update_task_definition(td, kwargs['force_td'])
    arn_s = ecs.install_service(ts, arn_td)

    # Echo ARNS
    click.secho("Task-definition ARN: {}".format(arn_td))
    click.secho("Service ARN: {}".format(arn_s))

    # Kill all currently running tasks
    if kwargs['fast_redeploy']:
        ecs.kill_tasks(ts.get('cluster'), arn_s)

    # Waiter
    if kwargs['wait_for_service']:
        client = boto3.client('ecs')
        waiter = client.get_waiter('services_stable')

        click.secho('Waiting until service is stable.')
        for i in range(5):
            try:
                waiter.wait(cluster=ts['cluster'], services=[service_name])
            except botocore.exceptions.WaiterError as e:
                if "Max attempts exceeded" in e.message:
                    click.secho("Service wasn't started in 600s")
                    continue
                click.secho(e.message)


if __name__ == '__main__':
    group()


import boto3


# Compare td and des['taskDefinition']
def _compare_dicts(incumbent, updateting):

    # This variable defines how to match members of a list while default id is name
    matching_pattern = {
        'portMappings': 'containerPort',
        'environmentFiles': 'type',
        'mountPoints': 'sourceVolume',
        'volumesFrom': 'sourceContainer',
        'devices': 'hostPath',
        'tmpfs': 'containerPath',
        'dependsOn': 'containerName',
        'extraHosts': 'hostname',
        'systemControls': 'namespace',
        'resourceRequirements': 'value',
        'placementConstraints': 'type',
        'inferenceAccelerators': 'deviceName',
        'tags': 'key'
    }


    # Get all keys
    current_keys = incumbent.keys()

    # Compare updating items with current values
    for k,v in updateting.items():

        # Simple test on key level
        if not k in current_keys:
            return False
        
        # k is even in the current, so lets compare values
        if v.__class__ == dict:

            # Only compare if both are dicts
            if incumbent[k].__class__ != dict:
                return False

            # Compare two nested dicts
            r = _compare_dicts(incumbent[k], v)
            if not r: return False
            continue
        
        # Compare two list
        if v.__class__ == list:
            # Only compare if both are lists
            if incumbent[k].__class__ != list:
                return False
            
            # Different lenght of lists will cause td update
            if len(v) != len(incumbent[k]):
                return False

            # Compare plain lists
            if v[0].__class__ != dict:
                if v == incumbent[k]:
                    continue
                return False
            
            # Match dicts - this is brutal implementation, and should be improved over time because
            # there are some parts of task definition this implementation might not work perfect.
            matching_id = matching_pattern.get(k, "name")
            for i in filter(lambda x: matching_id in x.keys(), v):
                # Find matching item
                matching_items = list(filter(lambda x: matching_id in x.keys() and x[matching_id] == i[matching_id], incumbent[k]))
                if len(matching_items) != 1:
                    return False
                if not _compare_dicts(matching_items[0], i):
                    return False

            # Stop evaluation
            continue
        
        # Compare values of simple types
        if v != incumbent[k]:
            return False
    
    # All the items are the same
    return True

    
def install_or_update_task_definition(td, force_update=False):
    # Setup connection to ECS interface
    client = boto3.client('ecs')

    # Verify the TD already exits
    a = client.list_task_definition_families(familyPrefix=td['family'])
    update = td['family'] in a.get('families', [])

    # Verify td is different compared to the current task definition revision
    if not force_update and update:
        des = client.describe_task_definition(taskDefinition=td['family'])
        if _compare_dicts(des['taskDefinition'], td):
            return des['taskDefinition']['taskDefinitionArn']
        
    # Register a new task definition
    return client.register_task_definition(**td)['taskDefinition']['taskDefinitionArn']


def install_service(sd, td_arn):
    """ 
    Function for installation of service which deals with the rendered service 
    definition and 
    """
    
    # Setup connection to ECS interface
    client = boto3.client('ecs')

    # Identify that service is installed or not
    service_arns = client.list_services(cluster=sd['cluster'])
    services = list(map(lambda x:x.split('/')[-1], service_arns.get('serviceArns', [])))
    update = sd['serviceName'] in services

    # Update service definition
    sd['taskDefinition'] = td_arn

    if update:
        sd['service'] = sd['serviceName']
        del sd['serviceName']
        if 'tags' in sd: del sd['tags']
        return client.update_service(**sd)['service']['serviceArn']
    else: 
        return client.create_service(**sd)['service']['serviceArn']


def kill_tasks(cluster, service_arn):
    """
    Function for stopping all currently running tasks off given service
    """
    # Setup connection to ECS interface
    client = boto3.client('ecs')

    # List tasks for given service
    lt = client.list_tasks(cluster=cluster, serviceName=service_arn.split('/')[-1])
    for task_arn in lt.get('taskArns', []):
        client.stop_task(cluster=cluster, task=task_arn)


from setuptools import setup

setup(
    name='ecscontrol',
    version='1.0.0',
    license='MIT',
    description = 'Tool for automation of deployments to AWS ECS',
    author = 'Milos Kozak',
    author_email = 'milos.kozak@lejmr.com',
    url = 'https://github.com/lejmr/ecsctl',
    download_url = 'https://github.com/lejmr/ecsctl/archive/v_01.tar.gz',    
    packages=['ecs', 'ecs.bin'],
    install_requires=[
        'Click',
        'Jinja2',
        'python-magic',
        'python-slugify',
        'dateparser',
        'boto3',
        'PyYAML'
    ],
    entry_points={
        'console_scripts': [
            'ecsdeploy = ecs.bin.ecsctl:group',
        ],
    },
    classifiers=[
        'Development Status :: 3 - Alpha',      # Chose either "3 - Alpha", "4 - Beta" or "5 - Production/Stable" as the current state of your package
        'Intended Audience :: Developers',      # Define that your audience are developers
        'Topic :: Software Development :: Build Tools',
        'License :: OSI Approved :: MIT License',   # Again, pick a license
        'Programming Language :: Python :: 3',      #Specify which pyhton versions that you want to support
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],
)


import re
from slugify import slugify
import dateparser
import boto3
from .exceptions import InvalidArnFormat

def get_jira_id(bn, default):
    # https://community.atlassian.com/t5/Bitbucket-questions/Regex-pattern-to-match-JIRA-issue-key/qaq-p/233319
    m = re.search('(?:[A-Z]{1,10}-?)([A-Z]+-\d+)', bn)
    if m:
        return m.group(0)
    return default

def get_slug(name, max_length=20):
    return slugify(name)[:max_length]

def filter_dateparser(s):
    return dateparser.parse(s)

def filter_format(d, f='%Y-%m-%dT%H:%M:%S.000'):
    # print(d.strftime(f))
    # return "milos"
    
    return str(d.strftime(f))

def _decode_parameter_store_secret(client, arn):
    param = arn.split(':')[5].split('parameter')[1]
    resp = client.get_parameter(Name=param, WithDecryption=True)
    return resp.get("Parameter", {}).get("Value", None)

def decode_aws_secret(arn):
    
    # Kill if not arn
    if arn[:3] != 'arn':
        raise InvalidArnFormat
    
    # Get the 
    a = arn.split(':')
    region = a[3]
    if a[2] == 'ssm':
        client = boto3.client('ssm', region_name=region)
        return _decode_parameter_store_secret(client, arn)

    return arn